/* VC_AHR_ENV_101.c
 *  
 * 
 * Device check default values:
 * avrdude -p m328p -c avrisp2                   						; command
 * avrdude: Device signature = 0x1e950f   (probably m328p)	; default values for m328p
 * avrdude: safemode: Fuses OK (E:FF, H:D9, L:62) 					; default values for m328p
 * 
 * 
 * Fuse setings for 20MHz external Quarz:
 * - avrdude -p m328p -c avrisp2 -U lfuse:w:0xe7:m
 * - avrdude -p m328p -c avrisp2 -U hfuse:w:0xd9:m
 * 
 * 
 * compile and flash:
 * - avr-gcc -Wall -Os -mmcu=atmega328p VC_AHR_ENV_101.c -o VC_AHR_ENV_101.elf
 * - avr-objcopy -O ihex -j .text -j .data VC_AHR_ENV_101.elf VC_AHR_ENV_101.hex
 * - avrdude -p m328p -c avrisp2 -U flash:w:VC_AHR_ENV_101.hex
 * 
 * 
 * Hardware: m328p at 20Mhz external
 * Board AHR_ENV V-1.2.0-A / 2024_11 and 1.3.0-B / 2025_01
 * 							 
 * 
 * ================================================================================
 * 
 * Ports used:
 * PB2-PB5 SPI interface
 * 
 * 
 * PC1 Attack CV
 * PC2 Release CV
 * 
 * 
 * PD0 LED/Out Release
 * PD1 LED/Out Attack
 * PD2 INT 0 Trigger
 * PD3 INT 1 Trigger
 * PD6 LED/Out EOC 
 * PD7 Gate/In
 * 
 * =========================================================================================
 * 
 * Generates a attack/(hold)/release envelope on rising edge of gate. Retriggers on attack and release state. Holds on gate high. Attack and
 * release time is  manually and voltage controllable. Gate is polled every loop.
 * Attack and release timing is table driven. To get other timing you have to change the tables and recompile.
 * CV input parameters are read with timer0 interrupt ISR (TIMER0_OVF_vect). 
 * Attack-, release- and EOC outputs are high when on. Signaled with LED.
 * Trigger input generates a attack decay envelope. No retrigger.
 * CV output voltage signaled with LED:
 * 
 * xx.12.2022
 * Working
 * 
 * 01.06.2023
 * Adapted to updated hardware. Board AHR_ENV V-1.0.1-A / 2024_03 and 1.0.1-B / 2024_03
 * Routines for ISR(INT0_vect) and ISR(INT1_vect) implemented. Yet empty.
 * 
 * 11.03.2024
 * Second lookup table implemented for different timing in attack and release section
 * 
 * 13.03.2024
 * Source code copied from VC_DAD_ENV.c
 * Code changed and added for DAR  
 * 
 * 16.03.2024
 * Second lookup table implemented for different timing in attack and decay section
 * Tables moved to flash. (PROGMEM)
 * 
 * 25.04.2024
 * ISR(INT0_vect implemented
 * 
 * 18.01.2025
 * Adapted to hardware revision 1.3.0-B	
 * 
*/

#define config_bits 0x03		// Mode set for MCP4921; output gain selected 1x

#include <avr/io.h>
#include <avr/interrupt.h>
#include <math.h>
#include <stdbool.h>
#include <avr/pgmspace.h>

#include "lookup_attack_DAD_7_4_1mS_5S.c"				// Mapping input voltage CVA to attack increment. Determine the attack timing.
#include "lookup_release_DAD_7_4_10mS_10S.c"		// Mapping input voltage CVD to release increment. Determine the release timing.

volatile float attack_time;											// Must be float here because of the calculation (no casting)
volatile float release_time;										// Must be float here because of the calculation (no casting)
float envelope = 0.0;
volatile uint8_t channel = 0;										// Channel number to be read successive one every interrupt
const uint16_t attack_target = 4095;						// Maximum output voltage
const uint8_t release_target = 0;								// Minimum output voltage

																								// Adapt the three constants below to your needs or preferences
const uint16_t envelope_high_target = 3900;			// Upper limit for envelope. Affects the timing of attack high tail
																								// To shorten the high tail increase envelope_high_target
																								// Remember to adjust the output voltage
const uint8_t envelope_low_target = 20;					// Lower limit for envelope. Affects the timing of release low tail
																								// To shorten the tail increase envelope_low_target
bool trigger = false;														// Goes high on rising edge of trigger input


void init_timer0(void)						// Timer interrupt used for reading CV 
{
//	TCCR0B = (1<<CS01)|(1<<CS00);	// Timer clock prescaler /64
	TCCR0B = (1<<CS02)|(1<<CS00);		// Timer clock prescaler /1024 ; 20MHz:1024 = 19531,25Hz; fast enough!
	TIFR0 = 1<<TOV0;								// Set overflow flag
	TIMSK0 = 1<<TOIE0;							// Enable overflow interrupt
}

void init_ports(void)
{
	DDRD |= (1<<PD0);									// Output for release signal and delay signal LED. High during delay.
	DDRD |= (1<<PD1);									// Output for attack signal and attack signal LED. High during attack.
	DDRD &=~ (1 << PD2);							// set PD2 as input; INT1  
  DDRD &=~ (1 << PD3);							// set PD3 as input; INT0  
	DDRD |= (1<<PD6);									// Output for end signal and EOC signal LED. High when the envelope is completed.
	DDRD &= ~(1 << PD7);							// PD7 Gate input
	EIMSK |= (1<<INT0);			        	// enable INT0 ; Uncomment those two lines if you want to use INT0
  EICRA |= (1<<ISC00)|(1<<ISC01); 	// INT0 - Rising edge on INT0 generates an interrupt request	
//  EIMSK |= (1<<INT1);			        // enable INT1 ; Uncomment those two lines if you want to use INT1
//  EICRA |= (1<<ISC11);            // INT1 - Falling edge of INT1 generates interrupt request
}


void init_ADC() 
{
	
  uint16_t result;
 
  ADMUX |= (1 << REFS0);					// reference voltage on AVCC    
	ADMUX |= (0 << REFS1);					// reference voltage on AVCC
	ADMUX |= (1 << ADLAR);					// left adjust result, return only 8bits
  // Bit ADFR ("free running") in ADCSRA is already set to zero on startup; single conversion
  ADCSRA |= (1<<ADPS2);    				// Prescaler 20MHz / 16 = 1250kHz recommended: 50 - 200 kHz
																	// Faster conversion usable here because we use only 8bits
  ADCSRA |= (1<<ADEN);            // ADC activate
  ADCSRA |= (1<<ADSC);						// eine ADC-Wandlung; dummy conversion 
  while (ADCSRA & (1<<ADSC))			// Wait for end of conversion
	{         	
  }
  result = ADCW;									//ADCW must be read once (ignore compiler warning)
}

/* ADC single conversion */
uint8_t ADC_Read(uint8_t channel)
{
  ADMUX = (ADMUX & ~(0x1F)) | (channel & 0x1F);	// Select channel without interfering other bits
  ADCSRA |= (1<<ADSC);            							// "single conversion"
  while (ADCSRA & (1<<ADSC))										// Wait for completion
	{  
  }
  return (ADCH);																// Only 8bits are used        
}


void init_SPI(void)
	{
		DDRB |= (1<<PB5)|(1<<PB3)|(1<<PB2);			// SCK, MOSI and SS as outputs
    DDRB &= ~(1<<PB4);											// MISO as input
    SPCR |= (1<<MSTR);											// Set AVR m328p as Master
    SPCR |= (0<<SPR0)|(0<<SPR1);						// divide clock by 4
//		SPCR |= (1<<SPR0)|(0<<SPR1);						// divide clock by 16
//		SPCR |= (0<<SPR0)|(1<<SPR1);						// divide clock by 64
//		SPCR |= (1<<SPR0)|(1<<SPR1);						// divide clock by 128
		SPSR |= (1<<SPI2X);											// double SPI speed (SCK)
    SPCR |= (1<<SPE);												// Enable SPI
	}


void Transmit_SPI_Master(uint16_t Data)						// Send two bytes to MCP4921
{	
	PORTB &=~ (1<<PB2);    												// assert the slave select
	SPDR = ((Data>>8) & 0x0F)|(config_bits<<4);		// Start transmission, send high byte first
	while (!(SPSR & (1<<SPIF)));									// Wait for transmission complete
	SPDR = Data & 0xFF;														// send low byte next
	while (!(SPSR & (1<<SPIF)));									// Wait for transmission complete
	PORTB |= (1<<PB2);														// de-assert slave select
}


void update_inputs(uint8_t scan)									// read the CV input parameters. One every loop.
	{        
		uint8_t attack_CV;
		uint8_t release_CV;
		
		switch (scan)
		{
			case 1:
				attack_CV = ADC_Read(1);									// get the attack time CV
				attack_time = pgm_read_float_near(&(lookup_attack[attack_CV]));	// get the attack time increment
			break; 
			case 2:
				release_CV = ADC_Read(2);									// get the release time CV
				release_time = pgm_read_float_near(&(lookup_release[release_CV]));		// get the release time increment
			break; 
		}
  }

//	Uncomment if you want to use the trigger input INT0. Don't forget init_port()s.
ISR(INT0_vect)  																	// Trigger rising edge
{
	trigger = true;																	// Trigger set true on rising edge
}  


/* Uncomment if you want to use INT1. Don't forget init_ports().
ISR(INT1_vect)              // Gate in goes low
{
	Put your code here
}  
*/

ISR (TIMER0_OVF_vect)									// (Did not work when variable channel is declared inside ISR)
	{
		update_inputs(channel);						// Update control voltage on timer interrupt
		channel = channel + 1;						// Select next channel
		if (channel > 2)									// All channels read?
			channel = 1;										// Start over
	}
	
	
int main(void)
	{
		bool gate = false;								// Holds gate state
		bool busy = false;								// True while envelope is active 

		
		cli();
		init_ports();
		init_SPI();
		init_ADC();
		init_timer0();
		trigger = false;
		sei();
		Transmit_SPI_Master(0);						// Set output to zero on start
		PORTD |= (1 << PD6);							// PD6 EOC state high on power on
		
		while(1)
		{        		
			gate = (PIND & (1 << PD7));									// Read gate state every loop
				if (((gate == true) || (trigger == true)) && (busy == false))	// Initial condition for envelope start: Gate == high or trigger == high and ENV not in progress
			{
				busy = true;															// ENV in progress
				PORTD &= ~(1 << PD6);											// PD6 EOC state off
				
				PORTD |= (1 << PD1);											// PD1 LED and attack state on
				while ((envelope < envelope_high_target))//  && (gate == true))	// Termination condition for while:  Max ENV voltage
				{																					// Setting to 4000 cuts off the "long tail" 4080 gives more of a plateau in the top. 
					envelope=((1.0-attack_time)*attack_target+attack_time*envelope);	// implement difference equation: y(k) = (1 - alpha) * x(k) + alpha * y(k-1)
					Transmit_SPI_Master(round(envelope));		// Output current envelope voltage
					if (((PIND & (1 << PD7)) == false) && (trigger == false))				// Read gate state during attack
						break;																// Stop attack phase on releasing key; Gate falling edge
				}																					// Attack reached maximum 
				PORTD &= ~(1 << PD1);											// PD1 LED and attack state off
				if (trigger == false)											// We are not in trigger mode
				{
					gate = (PIND & (1 << PD7));							// Read gate state
					while (gate == true)										// Wait for gate release == AHR ENV
					gate = (PIND & (1 << PD7));							// Read gate state until release
				}
																									// Release starts here. Move on when gate == false
				PORTD |= (1 << PD0);											// PD0 LED and release state on
				while ((envelope > envelope_low_target) && (gate == false))		// Termination condition for while: Minimum ENV voltage
																									// Setting to 80 cuts off the "long tail" 20 or lower gives more of a tail in the bottom.
				{																					// But this gives a bottom voltage value of 5/4096*envelope_low_target
					envelope=((1.0-release_time)*release_target+release_time*envelope);	// implement difference equation: y(k) = (1 - alpha) * x(k) + alpha * y(k-1)
					Transmit_SPI_Master(round(envelope));		// Output current envelope voltage
					gate = (PIND & (1 << PD7));							// Read gate state during decay
					if ((gate == true) && (trigger == false))		// "Re-gate"; Starts new envelope at gate on 
					{
						busy = false;
//						envelope = 0.0;											// Uncomment if you want to start next envelope from zero after break; "re-gate"
						break;
					}
				}																					// Release reached minimum
				Transmit_SPI_Master(0);										// Forces the output to zero. Might be useful if you choose to cut the long tail on release.
				PORTD &= ~(1 << PD0);											// PD0 LED and release state off
			}	// End of if gate = high
			if ((gate == false) || (trigger == true))		// This point is reached after the ENV has reached the end without "re-gate"
			{			
				busy = false;															// Ready for new AHR ENV
				trigger = false;													// Ready for new AD ENV
				PORTD |= (1 << PD6);											// PD6 LED and EOC state on
			}								
		}  
		return(0);																		// You'll never reach this line
	}
